// Copyright 2022 Bitcoding
// See the License for the specific language governing permissions and
// limitations under the License.

package com.btcpiyush.ads.google_applovin_unity_ads;

import android.util.Log;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.facebook.ads.Ad;
import com.facebook.ads.AdError;
import com.facebook.ads.RewardedVideoAd;
import com.facebook.ads.RewardedVideoAdListener;
import com.google.android.gms.ads.LoadAdError;
import com.google.android.gms.ads.rewarded.RewardItem;
import java.lang.ref.WeakReference;

/**
 * A wrapper for {@link RewardedVideoAd}.
 */
class FlutterFacebookRewardedAd extends FlutterAd.FlutterOverlayAd {
    private static final String TAG = "FlutterRewardedAd";

    @NonNull
    private final AdInstanceManager manager;
    @NonNull
    private final String adUnitId;
    @NonNull
    private final FlutterAdLoader flutterAdLoader;
    @Nullable
    RewardedVideoAd rewardedAd;

    /**
     * Constructor for facebook Ad Request.
     */
    public FlutterFacebookRewardedAd(
            int adId,
            @NonNull AdInstanceManager manager,
            @NonNull String adUnitId,
            @NonNull FlutterAdLoader flutterAdLoader) {
        super(adId);
        this.manager = manager;
        this.adUnitId = adUnitId;
        this.flutterAdLoader = flutterAdLoader;
    }

    @Override
    void load() {
        final RewardedVideoAdListener adLoadCallback = new DelegatingFacebookRewardedCallback(this);
        rewardedAd = flutterAdLoader.loadFacebookRewarded(manager.getActivity(), adUnitId, adLoadCallback);
      }

    void onAdLoaded() {
        manager.onAdLoaded(adId, null);
    }

    void onAdFailedToLoad(@NonNull LoadAdError loadAdError) {
        manager.onAdFailedToLoad(adId, new FlutterLoadAdError(loadAdError));
    }

    @Override
    public void show() {
        if (rewardedAd == null) {
            Log.e(TAG, "Error showing rewarded - the rewarded ad wasn't loaded yet.");
            return;
        }
        rewardedAd.show();
    }

    @Override
    public void setImmersiveMode(boolean immersiveModeEnabled) {
        if (rewardedAd == null) {
            Log.e(
                    TAG, "Error setting immersive mode in rewarded ad - the rewarded ad wasn't loaded yet.");
            return;
        }

    }

    void onUserEarnedReward(@NonNull RewardItem rewardItem) {
        manager.onRewardedAdUserEarnedReward(
                adId, new FlutterRewardedAd.FlutterRewardItem(rewardItem.getAmount(), rewardItem.getType()));
    }

    @Override
    void dispose() {
        if (rewardedAd != null)
            rewardedAd.destroy();
        rewardedAd = null;
    }

    /**
     * This class delegates various rewarded ad callbacks to FlutterRewardedAd. Maintains a weak
     * reference to avoid memory leaks.
     */
    private static final class DelegatingFacebookRewardedCallback
            implements RewardedVideoAdListener {

        private final WeakReference<FlutterFacebookRewardedAd> delegate;

        DelegatingFacebookRewardedCallback(FlutterFacebookRewardedAd delegate) {
            this.delegate = new WeakReference<>(delegate);
        }

        @Override
        public void onRewardedVideoCompleted() {
            if (delegate.get() != null) {
                delegate.get().onUserEarnedReward(new RewardItem() {
                    @Override
                    public int getAmount() {
                        return 0;
                    }

                    @NonNull
                    @Override
                    public String getType() {
                        return "";
                    }
                });
            }
        }

        @Override
        public void onRewardedVideoClosed() {
            if (delegate.get() != null) {
                delegate.get().manager.onAdDismissedFullScreenContent(delegate.get().adId);
            }
        }

        @Override
        public void onError(Ad ad, AdError adError) {
            if (delegate.get() != null) {
                delegate.get().onAdFailedToLoad(new LoadAdError(adError.getErrorCode(), adError.getErrorMessage(), "facebook.com", null, null));
            }
        }

        @Override
        public void onAdLoaded(Ad ad) {
            if (delegate.get() != null) {
                delegate.get().onAdLoaded();
            }
        }

        @Override
        public void onAdClicked(Ad ad) {
            if (delegate.get() != null) {
                delegate.get().manager.onAdClicked(delegate.get().adId);
            }
        }

        @Override
        public void onLoggingImpression(Ad ad) {
            if (delegate.get() != null) {
                delegate.get().manager.onAdImpression(delegate.get().adId);
            }
        }
    }
}
