// Copyright 2022 Bitcoding
// See the License for the specific language governing permissions and
// limitations under the License.


package com.btcpiyush.ads.google_applovin_unity_ads.usermessagingplatform;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;

import com.google.android.ump.ConsentForm;

import java.io.ByteArrayOutputStream;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import io.flutter.plugin.common.StandardMessageCodec;

/** Codec for UMP SDK. */
class UserMessagingCodec extends StandardMessageCodec {

  private static final byte VALUE_CONSENT_REQUEST_PARAMETERS = (byte) 129;
  private static final byte VALUE_CONSENT_DEBUG_SETTINGS = (byte) 130;
  private static final byte VALUE_CONSENT_FORM = (byte) 131;

  private final Map<Integer, ConsentForm> consentFormMap;

  UserMessagingCodec() {
    consentFormMap = new HashMap<>();
  }

  @Override
  protected void writeValue(@NonNull ByteArrayOutputStream stream, @NonNull Object value) {
    if (value instanceof ConsentRequestParametersWrapper) {
      stream.write(VALUE_CONSENT_REQUEST_PARAMETERS);
      ConsentRequestParametersWrapper params = (ConsentRequestParametersWrapper) value;
      writeValue(stream, params.getTfuac());
      writeValue(stream, params.getDebugSettings());
    } else if (value instanceof ConsentDebugSettingsWrapper) {
      stream.write(VALUE_CONSENT_DEBUG_SETTINGS);
      ConsentDebugSettingsWrapper debugSettings = (ConsentDebugSettingsWrapper) value;
      writeValue(stream, debugSettings.getDebugGeography());
      writeValue(stream, debugSettings.getTestIdentifiers());
    } else if (value instanceof ConsentForm) {
      stream.write(VALUE_CONSENT_FORM);
      writeValue(stream, value.hashCode());
    } else {
      super.writeValue(stream, value);
    }
  }

  @Nullable
  private List<String> asList(@Nullable Object maybeList) {
    if (maybeList == null) {
      return null;
    }
    List<String> stringList = new ArrayList<>();
    if (maybeList instanceof List) {
      List<?> list = (List<?>) maybeList;
      for (Object obj : list) {
        if (obj instanceof String) {
          stringList.add((String) obj);
        }
      }
    }
    return stringList;
  }

  @Override
  protected Object readValueOfType(byte type, @NonNull ByteBuffer buffer) {
    switch (type) {
      case VALUE_CONSENT_REQUEST_PARAMETERS:
        {
          Boolean tfuac = (Boolean) readValueOfType(buffer.get(), buffer);
          ConsentDebugSettingsWrapper debugSettings =
              (ConsentDebugSettingsWrapper) readValueOfType(buffer.get(), buffer);
          return new ConsentRequestParametersWrapper(tfuac, debugSettings);
        }
      case VALUE_CONSENT_DEBUG_SETTINGS:
        {
          Integer debugGeoInt = (Integer) readValueOfType(buffer.get(), buffer);
          List<String> testIdentifiers = asList(readValueOfType(buffer.get(), buffer));
          return new ConsentDebugSettingsWrapper(debugGeoInt, testIdentifiers);
        }
      case VALUE_CONSENT_FORM:
        {
          Integer hash = (Integer) readValueOfType(buffer.get(), buffer);
          return consentFormMap.get(hash);
        }
      default:
        return super.readValueOfType(type, buffer);
    }
  }

  void trackConsentForm(ConsentForm form) {
    consentFormMap.put(form.hashCode(), form);
  }

  void disposeConsentForm(ConsentForm form) {
    consentFormMap.remove(form.hashCode());
  }
}
