import 'package:google_applovin_unity_ads/google_applovin_unity_ads.dart';

/// Utility class that manages loading and showing app open ads.
class AppOpenAdManager {
  /// Maximum duration allowed between loading and showing the ad.
  final Duration maxCacheDuration = Duration(hours: 1);

  static const String _GOOGLEAD = "google_ads";
  static const String _ADMANAGERAD = "ad_manager";
  static const String _APPLOVINAD = "app_lovin";

  /// Keep track of load time so we don't show an expired ad.
  DateTime? _appOpenLoadTime;

  AppOpenAd? _appOpenAd;
  AppLovinOpenAd? _appLovinOpenAd;
  bool _isShowingAd = false;

  /// Load an [AppOpenAd].
  void loadAd({int pos = 0, Function? callback, Function? onFailedCallback}) {
    final _settings = getAdsSettings();

    if (_settings == null ||
        (_settings.adSetting?.appVersionCode == appVersionCode) ||
        (_settings.openAdsSequence?.isEmpty ?? true) ||
        (getAdsSettings()?.isAllAds ?? false) == false ||
        ((_settings.openAdsSequence?.length ?? 0) <= pos)) {
      if (onFailedCallback != null) {
        onFailedCallback();
      }
      return;
    }

    switch (_settings.openAdsSequence?[pos]) {
      case _GOOGLEAD:
      case _ADMANAGERAD:
        final adUnitId = _settings.googleAds?.appOpenAdId ?? "";

        if (adUnitId.isEmpty || !(_settings.googleAds?.isOpenAds ?? false)) {
          loadAd(pos: pos + 1, callback: callback);
          return;
        }

        AppOpenAd.load(
          adUnitId: adUnitId,
          orientation: AppOpenAd.orientationPortrait,
          request: AdRequest(),
          adLoadCallback: AppOpenAdLoadCallback(
            onAdLoaded: (ad) {
              print('$ad loaded');
              _appOpenLoadTime = DateTime.now();
              _appOpenAd = ad;
              if (callback != null) {
                callback();
              }
            },
            onAdFailedToLoad: (error) {
              print('AppOpenAd failed to load: $error');
              _appOpenAd = null;
              loadAd(pos: pos + 1, callback: callback);
            },
          ),
        );
        break;

      case _APPLOVINAD:
        final adUnitId = _settings.appLovin?.openId ?? "";

        if (adUnitId.isEmpty || !(_settings.appLovin?.isOpenAds ?? false)) {
          loadAd(pos: pos + 1, callback: callback);
          return;
        }

        AppLovinOpenAd.load(
          adUnitId: adUnitId,
          adLoadCallback: AppLovinOpenAdLoadCallback(
            onAdLoaded: (ad) {
              print('$ad loaded');
              _appOpenLoadTime = DateTime.now();
              _appLovinOpenAd = ad;
              if (callback != null) {
                callback();
              }
            },
            onAdFailedToLoad: (error) {
              print('AppOpenAd failed to load: $error');
              _appLovinOpenAd = null;
              loadAd(pos: pos + 1, callback: callback);
            },
          ),
        );
        break;

      default:
        loadAd(pos: pos + 1, callback: callback);
        break;
    }
  }

  /// Whether an ad is available to be shown.
  bool get isAdAvailable {
    return _appOpenAd != null || _appLovinOpenAd != null;
  }

  /// Shows the ad, if one exists and is not already being shown.
  ///
  /// If the previously cached ad has expired, this just loads and caches a
  /// new ad.
  void showAdIfAvailable() {
    if ((getAdsSettings()?.isAllAds ?? false) == false) {
      return;
    }

    if (!isAdAvailable) {
      print('Tried to show ad before available.');
      loadAd(callback: () => showAdIfAvailable());
      return;
    }
    if (_isShowingAd) {
      print('Tried to show ad while already showing an ad.');
      return;
    }
    if (_appOpenLoadTime != null && DateTime.now().subtract(maxCacheDuration).isAfter(_appOpenLoadTime!)) {
      print('Maximum cache duration exceeded. Loading another ad.');
      _appOpenAd?.dispose();
      _appOpenAd = null;
      _appLovinOpenAd?.dispose();
      _appLovinOpenAd = null;
      loadAd(callback: () => showAdIfAvailable());
      return;
    }

    if (_appOpenAd != null) {
      // Set the fullScreenContentCallback and show the ad.
      _appOpenAd?.fullScreenContentCallback = FullScreenContentCallback(
        onAdShowedFullScreenContent: (ad) {
          _isShowingAd = true;
          print('$ad onAdShowedFullScreenContent');
        },
        onAdFailedToShowFullScreenContent: (ad, error) {
          print('$ad onAdFailedToShowFullScreenContent: $error');
          _isShowingAd = false;
          ad.dispose();
          _appOpenAd = null;
        },
        onAdDismissedFullScreenContent: (ad) {
          print('$ad onAdDismissedFullScreenContent');
          _isShowingAd = false;
          ad.dispose();
          _appOpenAd = null;
          loadAd();
        },
      );
      _appOpenAd?.show();
    }

    if (_appLovinOpenAd != null) {
      // Set the fullScreenContentCallback and show the ad.
      _appLovinOpenAd?.fullScreenContentCallback = FullScreenContentCallback(
        onAdShowedFullScreenContent: (ad) {
          _isShowingAd = true;
          print('$ad onAdShowedFullScreenContent');
        },
        onAdFailedToShowFullScreenContent: (ad, error) {
          print('$ad onAdFailedToShowFullScreenContent: $error');
          _isShowingAd = false;
          ad.dispose();
          _appLovinOpenAd = null;
        },
        onAdDismissedFullScreenContent: (ad) {
          print('$ad onAdDismissedFullScreenContent');
          _isShowingAd = false;
          ad.dispose();
          _appLovinOpenAd = null;
          loadAd();
        },
      );
      _appLovinOpenAd?.show();
    }

  }
}